@@ -21,8 +21,13 @@
 #endif
 
 #include <stdio.h>
+#include <stdint.h>
 #include <stdlib.h>
 #include <string.h>
+#include <pthread.h>
+#ifdef HAVE_SYSCONF
+#include <unistd.h>
+#endif
 
 #include <igloo/config.h>
 #include <igloo/error.h>
@@ -32,13 +37,18 @@
 
 #define MAX_RECURSIONS 32
 
+#define MARK_FREEING    0x80000000
+
+#define IS_IMMUTABLE_TYPE(x)    ((((ptrdiff_t)(x)->type_flags) & igloo_RO_TYPEFLAG_IMMUTABLE) ? true : false)
+#define IS_IMMUTABLE(x)         IS_IMMUTABLE_TYPE(igloo_RO__GETSTUB((x))->type)
+
 struct igloo_ro_object_group_tag {
     igloo_ro_full_t __parent;
     /* lock for igloo_ro_*() */
     igloo_rwlock_t lock;
+    igloo_rwlock_t lock_immutable;
     /* instance */
     igloo_ro_t instance;
-    bool is_default_group;
 };
 
 static igloo_error_t igloo_ro_full_new(igloo_ro_t self, const igloo_ro_type_t *type, va_list ap);
@@ -54,7 +64,8 @@ static const igloo_feature_t ** features_full[] = {
 };
 
 igloo_RO_OPEN_TYPE(igloo_ro_stub_t, igloo_ro_stub_t,
-        igloo_RO_TYPEDECL_NEW_NOOP()
+        igloo_RO_TYPEDECL_NEW_NOOP(),
+        igloo_RO_TYPEDECL_FLAGS(igloo_RO_TYPEFLAG_IMMUTABLE)
         );
 
 igloo_RO_OPEN_TYPE(igloo_ro_tiny_t, igloo_ro_stub_t,
@@ -74,11 +85,53 @@ igloo_RO_OPEN_TYPE(igloo_ro_object_group_t, igloo_ro_full_t,
 igloo_RO_ALIAS_TYPE(igloo_ro_t, igloo_ro_stub_t);
 
 /* ---[ Validation functions ]--- */
+static pthread_once_t get_pagesize_once = PTHREAD_ONCE_INIT;
+static bool get_pagesize_done = false;
+static size_t get_pagesize_value = 0;
+
+static void get_pagesize_func(void)
+{
+#ifdef HAVE_SYSCONF
+    long val;
+
+#ifdef _SC_PAGESIZE
+    val = sysconf(_SC_PAGESIZE);
+    if (val > 1)
+        get_pagesize_value = val;
+#endif
+
+#ifdef _SC_PAGE_SIZE
+    if (!get_pagesize_value) {
+        val = sysconf(_SC_PAGE_SIZE);
+        if (val > 1)
+            get_pagesize_value = val;
+    }
+#endif
+#endif
+
+    if (!get_pagesize_value) {
+        // best guess.
+        get_pagesize_value = 1024;
+    }
+
+    get_pagesize_done = true;
+}
+
+
+static inline igloo_ATTR_F_HOT igloo_ATTR_F_CONST size_t get_pagesize(void)
+{
+    if (!get_pagesize_done)
+        pthread_once(&get_pagesize_once, get_pagesize_func);
+
+    return get_pagesize_value;
+}
+
 static inline igloo_ATTR_F_HOT igloo_ATTR_F_CONST bool is_valid_object_or_type_pointer(const void *addr)
 {
-    intptr_t ptr = (intptr_t)addr;
+#ifdef HAVE_UINTPTR_T
+    uintptr_t ptr = (uintptr_t)addr;
 
-    if (ptr < 1024)
+    if (ptr < get_pagesize())
         return false;
 
 #ifdef igloo_ATTR_T_ALIGNED_EFFECTIVE
@@ -87,6 +140,9 @@ static inline igloo_ATTR_F_HOT igloo_ATTR_F_CONST bool is_valid_object_or_type_p
 #endif
 
     return true;
+#else
+    return addr != NULL ? true : false;
+#endif
 }
 
 static inline igloo_error_t check_type(const igloo_ro_type_t *type)
@@ -94,6 +150,7 @@ static inline igloo_error_t check_type(const igloo_ro_type_t *type)
     const igloo_ro_type_t *cur = type;
     int loop = 0;
     size_t max_size = type->type_length;
+    bool immutable = IS_IMMUTABLE_TYPE(type);
 
     if (!is_valid_object_or_type_pointer(type))
         return igloo_ERROR_FAULT;
@@ -110,6 +167,14 @@ static inline igloo_error_t check_type(const igloo_ro_type_t *type)
         if (size > max_size || size < sizeof(igloo_ro_stub_t))
             return igloo_ERROR_FAULT;
 
+        if (IS_IMMUTABLE_TYPE(cur) != immutable) {
+            if (cur != igloo_RO_GET_OWN_TYPE_BY_SYMBOL(igloo_ro_stub_t) &&
+                cur != igloo_RO_GET_OWN_TYPE_BY_SYMBOL(igloo_ro_tiny_t) &&
+                cur != igloo_RO_GET_OWN_TYPE_BY_SYMBOL(igloo_ro_full_t)) {
+                return igloo_ERROR_FAULT;
+            }
+        }
+
         max_size = size;
 
         loop++;
@@ -162,7 +227,7 @@ static inline igloo_ATTR_F_HOT igloo_ATTR_F_PURE bool is_object_strong(igloo_ro_
     return is_valid_object(object) && igloo_RO__GETSTUB(object)->refc ? true : false;
 }
 
-static inline igloo_ATTR_F_HOT igloo_ATTR_F_PURE bool is_object_of_type_strong(igloo_ro_t object, const igloo_ro_type_t *type)
+static inline igloo_ATTR_F_HOT igloo_ATTR_F_PURE bool is_object_of_type_and_strong(igloo_ro_t object, const igloo_ro_type_t *type)
 {
     return is_object_of_type(object, type) && is_object_strong(object) ? true : false;
 }
@@ -180,11 +245,11 @@ bool            igloo_RO_HAS_TYPE_raw(igloo_ro_t object, const igloo_ro_type_t *
 }
 bool            igloo_RO_IS_VALID_raw(igloo_ro_t object, const igloo_ro_type_t *type)
 {
-    return is_object_of_type_strong(object, type);
+    return is_object_of_type_and_strong(object, type);
 }
 igloo_ro_t      igloo_RO_TO_TYPE_raw(igloo_ro_t object, const igloo_ro_type_t *type)
 {
-    return is_object_of_type_strong(object, type) ? object : igloo_RO_NULL;
+    return is_object_of_type_and_strong(object, type) ? object : igloo_RO_NULL;
 }
 /* ---[ End of public validation functions ]--- */
 
@@ -216,6 +281,7 @@ static igloo_error_t igloo_ro_object_group_new(igloo_ro_t self, const igloo_ro_t
     (void)type, (void)ap;
 
     igloo_rwlock_init(&(group->lock));
+    igloo_rwlock_init(&(group->lock_immutable));
 
     if (stub->group) {
         error = igloo_ro_get_instance(self, &(group->instance));
@@ -223,6 +289,9 @@ static igloo_error_t igloo_ro_object_group_new(igloo_ro_t self, const igloo_ro_t
             return error;
 
         error = igloo_ro_unref(&(stub->group));
+        if (error != igloo_ERROR_NONE)
+            return error;
+        error = igloo_ro_ref(group, &(stub->group), igloo_ro_object_group_t);
         if (error != igloo_ERROR_NONE)
             return error;
         stub->group = group;
@@ -235,13 +304,10 @@ static void igloo_ro_object_group_free(igloo_ro_t self)
 {
     igloo_ro_object_group_t *group = igloo_ro_to_type(self, igloo_ro_object_group_t);
 
-    if (group->is_default_group) {
-        igloo_ro_weak_unref(&(group->instance));
-    } else {
-        igloo_ro_unref(&(group->instance));
-    }
+    igloo_ro_unref(&(group->instance));
 
     igloo_rwlock_destroy(&(group->lock));
+    igloo_rwlock_destroy(&(group->lock_immutable));
 }
 /* ---[ End of type specific functions ]--- */
 
@@ -250,7 +316,9 @@ static void igloo_ro_object_group_free(igloo_ro_t self)
 static inline igloo_ATTR_F_HOT void igloo_ro_ ## action ## lock (igloo_ro_t self, bool full) \
 { \
     if (full) { \
-        func(&(self.subtype__igloo_ro_full_t->lock)); \
+        func(&(((igloo_ro_full_t*)igloo_RO__GETSTUB(self))->lock)); \
+    } else if (IS_IMMUTABLE(self)) { \
+        func(&(igloo_RO__GETSTUB(self)->group->lock_immutable)); \
     } else { \
         func(&(igloo_RO__GETSTUB(self)->group->lock)); \
     } \
@@ -260,6 +328,16 @@ LOCK_FUNC(igloo_rwlock_rlock, r)
 LOCK_FUNC(igloo_rwlock_wlock, rw)
 LOCK_FUNC(igloo_rwlock_unlock, un)
 LOCK_FUNC(igloo_rwlock_destroy, destroy)
+
+static inline igloo_ATTR_F_HOT void igloo_ro_roptwlock(igloo_ro_t self, bool full)
+{
+    if (IS_IMMUTABLE(self)) {
+        igloo_ro_rlock(self, full);
+    } else {
+        igloo_ro_rwlock(self, full);
+    }
+}
+
 /* ---[ End of object locking functions ]--- */
 
 /* ---[ Object creation functions ]--- */
@@ -391,8 +469,9 @@ igloo_error_t   igloo_ro_ref_raw(igloo_ro_t self, igloo_ro_t *out, const igloo_r
 
 static void igloo_ro_ungroup(igloo_ro_stub_t *stub)
 {
-    if (stub->group != NULL && (igloo_ro_stub_t*)stub->group != stub)
-        igloo_ro_unref(&(stub->group));
+    igloo_ro_object_group_t *group = stub->group;
+    stub->group = NULL;
+    igloo_ro_unref(&(group));
 }
 
 static igloo_error_t   igloo_ro_destroy(igloo_ro_t *self, bool full)
@@ -419,6 +498,43 @@ static void igloo_ro_run_free(igloo_ro_stub_t *stub)
     } while (cur != cur->type_parent && (cur = cur->type_parent));
 }
 
+static inline bool needs_to_be_kept_alive_instance(igloo_ro_stub_t *instance)
+{
+    if (instance->refc == 0)
+        return false;
+
+    if (instance->group != NULL) {
+        igloo_ro_stub_t *group = igloo_RO__GETSTUB(instance->group);
+        if (group->refc != 2)
+            return true;
+    }
+
+    return instance->refc != 1;
+}
+
+static inline bool needs_to_be_kept_alive_group(igloo_ro_stub_t *stub)
+{
+    igloo_ro_object_group_t *group = (igloo_ro_object_group_t*)stub;
+    igloo_ro_stub_t *instance = igloo_RO__GETSTUB(group->instance);
+
+    if (instance && instance->group == group) {
+        return stub->refc != 2 || needs_to_be_kept_alive_instance(instance);
+    } else {
+        return stub->refc > 1;
+    }
+}
+
+static inline bool needs_to_be_kept_alive(igloo_ro_stub_t *stub)
+{
+    if (stub->type == igloo_RO_GET_TYPE_BY_SYMBOL(igloo_ro_object_group_t)) {
+        return needs_to_be_kept_alive_group(stub);
+    } else if (stub->type == igloo_instance_type) {
+        return needs_to_be_kept_alive_instance(stub);
+    }
+
+    return stub->refc > 1;
+}
+
 igloo_error_t   igloo_ro_unref_raw(igloo_ro_t *self)
 {
     igloo_ro_stub_t *stub;
@@ -439,19 +555,22 @@ igloo_error_t   igloo_ro_unref_raw(igloo_ro_t *self)
 
     igloo_ro_rwlock(*self, full);
 
-    if (stub->refc > 1) {
-        stub->refc--;
+    stub->refc--;
+
+    if (needs_to_be_kept_alive(stub)) {
         igloo_ro_unlock(*self, full);
         *self = igloo_RO_NULL;
         return igloo_ERROR_NONE;
     }
 
+    stub->refc |= MARK_FREEING;
+
     igloo_ro_run_free(stub);
 
     if (full)
         igloo_ro_ungroup(stub);
 
-    stub->refc--;
+    stub->refc -= MARK_FREEING;
 
     if (stub->wrefc) {
         igloo_ro_unlock(*self, full);
@@ -520,7 +639,7 @@ igloo_error_t   igloo_ro_ref_replace_raw(igloo_ro_t self, igloo_ro_t *out, const
         return igloo_ERROR_FAULT;
 
     if (!igloo_ro_is_null(self)) {
-        if (!is_object_of_type_strong(self, type))
+        if (!is_object_of_type_and_strong(self, type))
             return igloo_ERROR_INVAL;
 
         error = igloo_ro_ref_transparent(self, &ref);
@@ -545,7 +664,7 @@ igloo_error_t   igloo_ro_weak_ref_replace_raw(igloo_ro_t self, igloo_ro_t *out,
         return igloo_ERROR_FAULT;
 
     if (!igloo_ro_is_null(self)) {
-        if (!is_object_of_type_strong(self, type))
+        if (!is_object_of_type_and_strong(self, type))
             return igloo_ERROR_INVAL;
 
         error = igloo_ro_weak_ref_transparent(self, &ref);
@@ -618,7 +737,7 @@ igloo_error_t   igloo_ro_get_error_raw(igloo_ro_t self, igloo_error_t *result, c
     full = is_object_full(self);
     igloo_ro_rlock(self, full);
 
-    if (!is_object_of_type_strong(self, type)) {
+    if (!is_object_of_type_and_strong(self, type)) {
         igloo_ro_unlock(self, full);
         return igloo_ERROR_GONE;
     }
@@ -637,10 +756,12 @@ igloo_error_t   igloo_ro_get_error_raw(igloo_ro_t self, igloo_error_t *result, c
 
 igloo_error_t   igloo_ro_stringify_raw(igloo_ro_t self, char **result, igloo_ro_sy_t flags, const igloo_ro_type_t *type)
 {
-#define STRINGIFY_FORMAT_STRONG     "{%p@%s, strong, %s, %zu static bytes, group={%p@%s}}", \
-    stub, igloo_RO_GET_TYPENAME(self), (full ? "full" : "tiny"), igloo_RO_GET_TYPE(self)->type_length, stub->group, igloo_RO_GET_TYPENAME(stub->group)
-#define STRINGIFY_FORMAT_WEAK       "{%p@%s, weak, %s, %zu static bytes}", \
-    stub, igloo_RO_GET_TYPENAME(self), (full ? "full" : "tiny"), igloo_RO_GET_TYPE(self)->type_length
+#define STRINGIFY_FORMAT_STRONG     "{%p@%s, strong (%" PRIu32 "/%" PRIu32 "), %s, %s, %zu static bytes, group={%p@%s}}", \
+    stub, igloo_RO_GET_TYPENAME(self), stub->refc, stub->wrefc, (full ? "full" : "tiny"), (immutable ? "immutable" : "mutable"), \
+    igloo_RO_GET_TYPE(self)->type_length, stub->group, igloo_RO_GET_TYPENAME(stub->group)
+#define STRINGIFY_FORMAT_WEAK       "{%p@%s, weak (%" PRIu32 "/%" PRIu32 "), %s, %s, %zu static bytes}", \
+    stub, igloo_RO_GET_TYPENAME(self), stub->refc, stub->wrefc, (full ? "full" : "tiny"), (immutable ? "immutable" : "mutable"), \
+    igloo_RO_GET_TYPE(self)->type_length
 #define STRINGIFY_FORMAT_INVALID    "{%p, invalid}", \
     igloo_RO__GETSTUB(self)
 #define STRINGIFY_TEXT_NULL     "{igloo_RO_NULL}"
@@ -664,10 +785,11 @@ igloo_error_t   igloo_ro_stringify_raw(igloo_ro_t self, char **result, igloo_ro_
         } \
     } while (0)
 
-    igloo_ro_stub_t *stub = igloo_ro_to_type(self, igloo_ro_stub_t);
+    igloo_ro_stub_t *stub = is_object_of_type(self, igloo_RO_GET_TYPE_BY_SYMBOL(igloo_ro_stub_t)) ? igloo_RO__GETSTUB(self) : NULL;
     igloo_error_t error = igloo_ERROR_NOSYS;
     igloo_ro_stringify_t impl = NULL;
     bool full;
+    bool immutable;
 
     if (!result)
         return igloo_ERROR_FAULT;
@@ -691,8 +813,12 @@ igloo_error_t   igloo_ro_stringify_raw(igloo_ro_t self, char **result, igloo_ro_
 
     // check for invalid pointers: when igloo_ro_is_null(self) is not true but stub == NULL we must have an invalid pointer.
     if (!stub) {
-        STRINGIFY_PROC(STRINGIFY_FORMAT_INVALID);
-        return error;
+        if (flags & igloo_RO_SY_OBJECT) {
+            STRINGIFY_PROC(STRINGIFY_FORMAT_INVALID);
+            return error;
+        } else {
+            return igloo_ERROR_FAULT;
+        }
     }
 
     if (!type)
@@ -702,15 +828,20 @@ igloo_error_t   igloo_ro_stringify_raw(igloo_ro_t self, char **result, igloo_ro_
         return igloo_ERROR_TYPEMM;
 
     full = is_object_full(self);
+    immutable = IS_IMMUTABLE(self);
 
     if (!stub->refc) {
-        STRINGIFY_PROC(STRINGIFY_FORMAT_WEAK);
-        return error;
+        if (flags & igloo_RO_SY_OBJECT) {
+            STRINGIFY_PROC(STRINGIFY_FORMAT_WEAK);
+            return error;
+        } else {
+            return igloo_ERROR_GONE;
+        }
     }
 
-    igloo_ro_rwlock(self, full);
+    igloo_ro_roptwlock(self, full);
 
-    if (!is_object_of_type_strong(self, type)) {
+    if (!is_object_of_type_and_strong(self, type)) {
         igloo_ro_unlock(self, full);
         return igloo_ERROR_GONE;
     }
@@ -749,7 +880,7 @@ bool            igloo_ro_can_raw(igloo_ro_t self, igloo_feature_t *feature, cons
     full = is_object_full(self);
     igloo_ro_rlock(self, full);
 
-    if (!is_object_of_type_strong(self, type)) {
+    if (!is_object_of_type_and_strong(self, type)) {
         igloo_ro_unlock(self, full);
         return false;
     }
@@ -811,8 +942,11 @@ igloo_error_t igloo_ro_bootstrap(igloo_ro_t *instance, const igloo_ro_type_t *ty
     if (error != igloo_ERROR_NONE)
         return error;
 
-    igloo_RO__GETSTUB(group)->group = group;
-    group->is_default_group = true;
+    error = igloo_ro_ref(group, &(igloo_RO__GETSTUB(group)->group), igloo_ro_object_group_t);
+    if (error != igloo_ERROR_NONE) {
+        igloo_ro_unref(&group);
+        return error;
+    }
 
     error = igloo_ro_new__raw_checked(instance, type, group);
     if (error != igloo_ERROR_NONE) {
@@ -820,7 +954,7 @@ igloo_error_t igloo_ro_bootstrap(igloo_ro_t *instance, const igloo_ro_type_t *ty
         return error;
     }
 
-    error = igloo_ro_weak_ref_transparent(*instance, &(group->instance));
+    error = igloo_ro_ref_transparent(*instance, &(group->instance));
     if (error != igloo_ERROR_NONE) {
         igloo_ro_unref(&group);
         igloo_ro_unref(instance);
